/*
 * Copyright (C) 2006-2013 Christopho, Solarus - http://www.solarus-games.org
 * 
 * Solarus is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Solarus is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#include "entities/Switch.h"
#include "entities/Arrow.h"
#include "entities/Block.h"
#include "entities/Hero.h"
#include "lua/LuaContext.h"
#include "Sprite.h"
#include "Game.h"
#include "Map.h"
#include "lowlevel/FileTools.h"
#include "lowlevel/Debug.h"
#include "lowlevel/StringConcat.h"
#include "lowlevel/Sound.h"

/**
 * \brief Constructor.
 * \param name name of the entity
 * \param layer layer of the entity
 * \param x x position of the entity's rectangle
 * \param y y position of the entity's rectangle
 * \param subtype the subtype of switch
 * \param needs_block true if a block is required to activate this switch
 * \param inactivate_when_leaving true to inactivate the switch when the hero or
 * the block leaves it
 */
Switch::Switch(
    const std::string& name,
    Layer layer,
    int x,
    int y,
    Subtype subtype,
    bool needs_block,
    bool inactivate_when_leaving):
  Detector(COLLISION_NONE, name, layer, x, y, 16, 16),
  subtype(subtype),
  activated(false),
  locked(false),
  needs_block(needs_block),
  inactivate_when_leaving(inactivate_when_leaving),
  entity_overlapping(NULL) {

  // sprite
  if (is_walkable()) {
    set_collision_modes(COLLISION_CUSTOM);
    if (subtype == WALKABLE_VISIBLE) {
      create_sprite("entities/switch");
      get_sprite().set_current_animation("inactivated");
    }
  }
  else if (subtype == ARROW_TARGET) {
    set_collision_modes(COLLISION_FACING_POINT);
  }
  else if (subtype == SOLID) {
    create_sprite("entities/solid_switch");
    get_sprite().set_current_animation("inactivated");
    set_collision_modes(COLLISION_SPRITE | COLLISION_RECTANGLE);
    set_optimization_distance(2000);  // Because of bombs and arrows on the switch.
  }
}

/**
 * \brief Destructor.
 */
Switch::~Switch() {

}

/**
 * \brief Returns the type of entity.
 * \return the type of entity
 */
EntityType Switch::get_type() const {
  return SWITCH;
}

/**
 * \brief Returns whether this entity is an obstacle for another one when
 * it is enabled.
 * \param other another entity
 * \return true if this entity is an obstacle for the other one
 */
bool Switch::is_obstacle_for(MapEntity& other) {
  return other.is_switch_obstacle(*this);
}

/**
 * \brief Returns whether this switch is a walkable switch.
 * \return true if the subtype of this switch is WALKABLE_INVISIBLE or WALKABLE_VISIBLE
 */
bool Switch::is_walkable() {
  return subtype == WALKABLE_INVISIBLE || subtype == WALKABLE_VISIBLE;
}

/**
 * \brief Returns whether this switch is an arrow target.
 * \return true if the subtype of this switch is ARROW_TARGET
 */
bool Switch::is_arrow_target() {
  return subtype == ARROW_TARGET;
}

/**
 * \brief Returns whether this switch is a solid switch.
 * \return true if the subtype of this switch is SOLID
 */
bool Switch::is_solid() {
  return subtype == SOLID;
}

/**
 * \brief Returns whether this switch is currently activated.
 *
 * \return true if the switch is activated
 */
bool Switch::is_activated() {
  return activated;
}

/**
 * \brief Activates the switch, playing a sound and notifying the map script.
 *
 * This function does nothing if the switch is locked or already activated.
 */
void Switch::activate() {

  if (!activated && !locked) {

    set_activated(true);

    if (subtype == WALKABLE_VISIBLE || subtype == SOLID) {
      Sound::play("switch");
    }

    get_lua_context().switch_on_activated(*this);
  }
}

/**
 * \brief Activates or inactivates the switch, not playing any sound.
 *
 * This function can change the switch state even if the switch is locked.
 *
 * \param activated true to make the switch on, false to make it off
 */
void Switch::set_activated(bool activated) {

  if (activated != this->activated) {
    this->activated = activated;

    if (subtype == WALKABLE_VISIBLE || subtype == SOLID) {
      if (activated) {
        get_sprite().set_current_animation("activated");
      }
      else {
        get_sprite().set_current_animation("inactivated");
      }
    }
  }
}

/**
 * \brief Locks this switch is its current state or unlocks it.
 *
 * When the switch is locked, it cannot be activated or inactivated by other entities.
 * However, the state can still be changed manually by calling set_activated().
 *
 * \param locked true to lock the switch in its current state, false to unlock it
 */
void Switch::set_locked(bool locked) {
  this->locked = locked;
}

/**
 * \brief Updates this switch.
 */
void Switch::update() {

  Detector::update();

  if (is_enabled() && is_walkable() && entity_overlapping != NULL) {

    // if an entity was on the switch, see if it is still there
    entity_overlapping_still_present = false;
    check_collision(*entity_overlapping);

    if (!entity_overlapping_still_present) {
      // the entity just left the switch or disappeared from the map
      // (it may even have been freed)

      entity_overlapping = NULL;
      if (is_activated() && inactivate_when_leaving && !locked) {
        set_activated(false);
        get_lua_context().switch_on_inactivated(*this);
      }
      get_lua_context().switch_on_left(*this);
    }
  }
}

/**
 * \brief Tests whether an entity's collides with this entity.
 * \param entity an entity
 * \return true if the entity's collides with this entity
 */
bool Switch::test_collision_custom(MapEntity& entity) {

  // this collision test is performed for walkable switches only

  const Rectangle& entity_rectangle = entity.get_bounding_box();
  int x1 = entity_rectangle.get_x() + 4;
  int x2 = x1 + entity_rectangle.get_width() - 9;
  int y1 = entity_rectangle.get_y() + 4;
  int y2 = y1 + entity_rectangle.get_height() - 9;

  return overlaps(x1, y1) && overlaps(x2, y1) &&
    overlaps(x1, y2) && overlaps(x2, y2);
}

/**
 * \brief This function is called by the engine when an entity overlaps the switch.
 * \param entity_overlapping the entity overlapping the detector
 * \param collision_mode the collision mode that detected the collision
 */
void Switch::notify_collision(MapEntity& entity_overlapping, CollisionMode collision_mode) {

  if (&entity_overlapping == this->entity_overlapping) {
    // already overlapping
    entity_overlapping_still_present = true;
    return;
  }

  if (!locked) {
    entity_overlapping.notify_collision_with_switch(*this, collision_mode);
  }
}

/**
 * \brief Notifies this entity that another sprite is overlapping it.
 *
 * This function is called by check_collision(MapEntity*, Sprite*) when another entity's
 * sprite overlaps a sprite of this detector.
 *
 * \param other_entity the entity overlapping this detector
 * \param other_sprite the sprite of other_entity that is overlapping this detector
 * \param this_sprite the sprite of this detector that is overlapping the other entity's sprite
 */
void Switch::notify_collision(MapEntity& other_entity, Sprite& other_sprite, Sprite& this_sprite) {

  if (!locked) {
    other_entity.notify_collision_with_switch(*this, other_sprite);
  }
}

/**
 * \brief This function is called when the hero overlaps this switch.
 *
 * The switch is activated if its properties allow it.
 *
 * \param hero the hero
 */
void Switch::try_activate(Hero& hero) {

  if (is_walkable() && !needs_block && !is_activated()) {
    // this switch allows the hero to activate it
    activate();
  }
  this->entity_overlapping = &hero;
}

/**
 * \brief This function is called when a block overlaps this switch.
 *
 * The switch is activated if its properties allow it.
 *
 * \param block the block overlapping this switch
 */
void Switch::try_activate(Block& block) {

  if (subtype == WALKABLE_VISIBLE && !is_activated()) {
    // blocks can only activate walkable, visible switches
    activate();
  }
  this->entity_overlapping = &block;
}

/**
 * \brief This function is called when an arrow overlaps this switch.
 *
 * The switch is activated if its properties allow it.
 *
 * \param arrow the arrow overlapping this switch
 */
void Switch::try_activate(Arrow& arrow) {

  if ((subtype == ARROW_TARGET || subtype == SOLID)
      && !is_activated()) {
    // arrows can only activate arrow targets and solid switches
    activate();
  }
}

/**
 * \brief This function is called when any entity overlaps this switch.
 *
 * Only solid switches can be activated this way.
 * The switch is activated if its properties allow it.
 */
void Switch::try_activate() {

  // arbitrary entities can activate solid switches
  if (subtype == SOLID && !is_activated()) {
    activate();
  }
}

/**
 * \brief Returns the name identifying this type in Lua.
 * \return The name identifying this type in Lua.
 */
const std::string& Switch::get_lua_type_name() const {
  return LuaContext::entity_switch_module_name;
}

